# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CLI for Logstory."""

import datetime
import glob
import os
import sys
import uuid

from absl import app
from absl import flags
from google.oauth2 import service_account
from google.cloud import storage

UTC = datetime.timezone.utc


# Validation function to check if the string is a valid UUID4
def is_valid_uuid4(value):
  """Abseil flag validation for customer id."""
  try:
    val = uuid.UUID(value, version=4)
    # Check if the string conforms to the UUID4 format
    return str(val) == value
  except ValueError:
    return False


def validate_service_account_json(json_path):
  """Abseil flag validation for credentials file."""
  # Check if the file exists
  if not os.path.isfile(json_path):
    raise FileNotFoundError(
        f"File does not exist: {json_path}.\n"
        "Please provide the complete path to a JSON credentials file.\n"
    )
  try:
    # Attempt to load the service account credentials
    _ = service_account.Credentials.from_service_account_file(json_path)
    # If it loads successfully, the path is valid
    return True, "The JSON file is valid."
  except Exception as e:
    # If an error occurs, return False and the error message
    # pylint: disable-next=broad-exception-raised
    raise Exception(f"The JSON file is invalid: {e}") from e


FLAGS = flags.FLAGS
flags.DEFINE_string(
    "credentials_path",
    None,
    "Path to JSON credentials for Ingestion API Service account.",
    short_name="c",
    required=False,  # conditionally req. depending on subcommand
)
flags.DEFINE_string(
    "customer_id",
    None,
    "Customer ID for SecOps instance, found on `/settings/profile/`.",
    required=False,  # conditionally req. depending on subcommand
)
flags.DEFINE_string(
    "region",
    None,  # "US" (default) is a regionless base url, so None works here.
    "SecOps tenant's region (Default=US). Used to set ingestion API base URL.",
    short_name="r",
    required=False,
)
flags.DEFINE_bool("entities", False, "Load Entities instead of Events")
flags.DEFINE_bool("three_day", False, "Load Entities instead of Events")
flags.DEFINE_string(
    "timestamp_delta",
    None,
    "Determines how datetimes in logfiles are updated. "
    "It is expressed in any/all: days, hours, minutes (d, h, m) (Default=1d) "
    "Examples: [1d, 1d1h, 1h1m, 1d1m, 1d1h1m, 1m1h, ...] "
    "Setting only `Nd` preserves the original HH:MM:SS but updates date. "
    "Nh/Nm subtracts an additional offset from that datetime, to facilitate "
    "running logstory more than 1x per day. ",
    required=False,
)
flags.DEFINE_string(
    "usecases_bucket",
    "logstory-usecases-20241216",  # default public bucket with usecases
    "GCP Cloud Storage bucket with additional usecases.",
    required=False,
)


def _get_current_time():
  """Returns the current time in UTC."""
  return datetime.datetime.now(UTC)


def usecases_list_logtypes():
  usecases_list(print_logtypes=True)


def get_usecases():
  """NOTE: this is *all* usecases, not just enabled in `events-24h`"""
  usecase_dirs = glob.glob(
      os.path.join(os.path.dirname(os.path.abspath(__file__)), "usecases/*")
  )
  usecases = []
  for usecase_dir in usecase_dirs:
    parts = os.path.split(usecase_dir)
    usecases.append(parts[-1])
  return usecases


def _get_blobs(bucket_name, usecase=None):
  client = storage.Client.create_anonymous_client()
  bucket = client.bucket(bucket_name)
  if usecase:
    blobs = bucket.list_blobs(prefix=usecase)
  else:
    blobs = bucket.list_blobs(delimiter="/")
  return blobs


def list_bucket_directories(bucket_name):
  """Lists the top-level directories in a GCP bucket."""
  blobs = _get_blobs(bucket_name)
  top_level_directories = []
  print(f"\nAvailable usecases in bucket '{bucket_name}':")
  for blob in blobs.pages:
    prefixes = blob.prefixes
    for prefix in prefixes:
      if "docs" in prefix:
        continue
      prefix = prefix.strip("/")
      print(f"- {prefix}")
      top_level_directories.append(prefix)
  return top_level_directories


def usecase_get(bucket_name, usecase):
  """Download usecase from GCP bucket."""
  blob_list = _get_blobs(bucket_name, usecase)
  for blob in blob_list:
    if blob.name.endswith("/"):
      continue
    destination_file_name = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "usecases/", blob.name
    )
    os.makedirs(os.path.dirname(destination_file_name), exist_ok=True)
    print(f"Downloading {blob.name} to {destination_file_name}")
    blob.download_to_filename(destination_file_name)


# Define subcommands
def usecases_list(print_logtypes=False, entities=False):
  """Walk the usecases/ dir and print the dir names."""
  entity_or_event = "ENTITIES" if entities else "EVENTS"
  usecase_dirs = glob.glob(
      os.path.join(os.path.dirname(os.path.abspath(__file__)), "usecases/*")
  )
  usecases = []
  logypes_map = {}
  markdown_map = {}
  for usecase_dir in usecase_dirs:
    parts = os.path.split(usecase_dir)
    if parts[-1] in ["__init__.py", "AWS"]:
      continue
    usecases.append(parts[-1])
    markdown_map[usecases[-1]] = []
    for md in glob.glob(os.path.join("./", usecase_dir, "*.md")):
      markdown_map[usecases[-1]].append(md)
    log_types = []
    if print_logtypes:
      for adir in glob.glob(os.path.join("./", usecase_dir, entity_or_event, "*.log")):
        log_types.append(os.path.splitext(os.path.split(adir)[-1])[0])
      logypes_map[usecases[-1]] = log_types
  for usecase in sorted(usecases):
    print(f"#\n# {usecase}\n#")
    for md in markdown_map.get(usecase, []):
      with open(md) as fh:
        print(fh.read())
    if print_logtypes:
      for log_type in sorted(logypes_map[usecase]):
        print(f"\t{log_type}")


def get_logtypes(usecase, entities=False):
  """Walk the subdirs in usecases and print the file base names."""
  print(f"Executing usecase: {usecase}")
  # Access common flags with FLAGS.flag_name

  entity_or_event = "ENTITIES" if entities else "EVENTS"
  usecase_dir = f"{os.path.split(__file__)[0]}/usecases/{usecase}/{entity_or_event}/"
  log_files = glob.glob(usecase_dir + "*.log")
  log_types = []
  for log_file in log_files:
    print(f"log_file: {log_file}")
    parts = os.path.split(log_file)
    log_type = os.path.splitext(parts[-1])[0]
    log_types.append(log_type)
    print(f"log_type: {log_type}")
  return log_types


def local_main(argv):
  """Parse args and call the appropriate function."""
  if len(argv) < 2:
    custom_help()
    return 1
  # parse the flags
  FLAGS(argv)
  # Some params are conditionally required
  # Check if either "usecase_replay" or "usecase_replay_logtype" is in argv
  requires_customer_and_credentials = any(
      arg in argv for arg in ["usecase_replay", "usecase_replay_logtype"]
  )
  # Check if customer_id or credentials_path is missing
  missing_customer_or_credentials = not FLAGS.customer_id or not FLAGS.credentials_path
  # Combine the conditions
  if requires_customer_and_credentials and missing_customer_or_credentials:
    # Handle the case where either customer_id or credential_path is missing
    raise ValueError(
        "The following flags must be set when using the usecase_replay and"
        " usecases_list_logtypes subcommands:\n\t--credentials_path,"
        " --customer_id, ..."
    )

  # optionality was determined above. now validate if the params are present.
  if FLAGS.credentials_path:
    flags.register_validator(
        "credentials_path",
        validate_service_account_json,
        message=("--credentials_path must be JSON credentials for a Service Account"),
    )
  if FLAGS.customer_id:
    flags.register_validator(
        "customer_id",
        is_valid_uuid4,
        message="--customer_id must be a UUID4",
    )
  # must parse flags again to enforce validators
  FLAGS(argv)

  # if present and valid, set as env vars
  if FLAGS.customer_id:
    os.environ["CUSTOMER_ID"] = FLAGS.customer_id
    print(f"The customer_id is: {os.environ['CUSTOMER_ID']}")

  if FLAGS.credentials_path:
    os.environ["CREDENTIALS_PATH"] = FLAGS.credentials_path
    print(f"The credentials_path is: {os.environ['CREDENTIALS_PATH']}")

  # should us -> '' logic be here or in the main.py file?
  if FLAGS.region:
    os.environ["REGION"] = FLAGS.region

  if len(argv) < 2:
    print("Usage: %s {usecases_list, command_two} [options]" % argv[0])
    return 1
  command = argv[1]
  if command == "usecases_list":
    return usecases_list()
  if command == "usecases_list_logtypes":
    return usecases_list_logtypes()
  if command == "usecases_list_available":
    return list_bucket_directories(FLAGS.usecases_bucket)
  if command == "usecase_get":
    usecase = argv[2]  # flaky?
    if not usecase:
      raise ValueError("Missing usecase argument for usecase_get")
    if usecase not in list_bucket_directories(FLAGS.usecases_bucket):
      raise ValueError(
          f"Usecase '{usecase}' not found in bucket '{FLAGS.usecases_bucket}'"
      )

    return usecase_get(FLAGS.usecases_bucket, usecase)

  # late import here due to globals in the module
  try:
    from . import main as imported_main
  except ImportError:
    import main as imported_main

  if command == "usecases_replay":
    usecases = get_usecases()
    logtypes = "*"
  elif command == "usecase_replay":
    usecase = argv[2]  # flaky?
    usecases = [
        usecase,
    ]
    logtypes = get_logtypes(usecase, entities=FLAGS.entities)
  elif command == "usecase_replay_logtype":
    usecase = argv[2]  # flaky?
    usecases = [
        usecase,
    ]
    logtypes = argv[3].split(",")
  else:
    print("Unknown command: %s" % command)
    return 1

  logstory_exe_time = datetime.datetime.now(datetime.timezone.utc)
  for use_case in usecases:
    logstory_exe_time = _get_current_time()
    if command == "usecases_replay":
      logtypes = get_logtypes(use_case)

    old_base_time = None  # resets for each usecase
    for log_type in logtypes:
      old_base_time = None  # resets for each log_type
      log_type = log_type.strip()
      print(f"usecase: {use_case}")
      print(f"{logtypes}, |{log_type}|")
      old_base_time = imported_main.usecase_replay_logtype(
          use_case,
          log_type,
          logstory_exe_time,
          old_base_time,
          timestamp_delta=FLAGS.timestamp_delta or None,
          entities=FLAGS.entities,
      )
    print(
        f"""UDM Search for the loaded logs:
    metadata.ingested_timestamp.seconds >= {int(logstory_exe_time.timestamp())}
    metadata.ingestion_labels["log_replay"]="true"
    metadata.ingestion_labels["replayed_from"]="logstory"
    metadata.ingestion_labels["usecase_name"]="{use_case}"
    """  # ToDo: get upper bound of ingested and print it?
    )

  return "Success Events!"


def custom_help():
  help_text = """
    Usage: logstory <command> [command options]

    Commands:
      usecases_list            List the usecases

      usecases_list_logtypes   List each logtype for each usecase

      usecases_list_available  List usecases that can be downloaded

      usecase_get              Download usecase from GCP bucket

      usecases_replay          Replay all usecases

      usecase_replay           Replay a specific usecase
         examples:
           usecase_replay NETWORK_ANALYSIS
           usecase_replay RULES_SEARCH_WORKSHOP
           ...

      usecase_replay_logtype   Replay a specific usecase with a specific log type
         examples:
           usecase_replay_logtype RULES_SEARCH_WORKSHOP POWERSHELL
           usecase_replay_logtype RULES_SEARCH_WORKSHOP WINEVETLOG
           usecase_replay_logtype RULES_SEARCH_WORKSHOP POWERSHELL,WINEVETLOG # No spaces without quotes
           usecase_replay_logtype RULES_SEARCH_WORKSHOP "POWERSHELL, WINEVETLOG" # Use quotes if there are spaces
           ...

      usecase_get EDR_WORKSHOP

    Usage Examples:
      logstory usecases_list

      logstory usecases_list_logtypes

      logstory usecases_replay --flagfile=config.cfg

      logstory usecase_replay NETWORK_ANALYSIS --flagfile=config.cfg
      logstory usecase_replay NETWORK_ANALYSIS --flagfile=config.cfg --timestamp_delta=1d
      logstory usecase_replay NETWORK_ANALYSIS --flagfile=config.cfg --timestamp_delta=1d1h
      logstory usecase_replay NETWORK_ANALYSIS --flagfile=config.cfg --timestamp_delta=1d1h1m
      logstory usecase_replay NETWORK_ANALYSIS --flagfile=config.cfg --timestamp_delta=1h30m

      logstory usecase_replay_logtype RULES_SEARCH_WORKSHOP POWERSHELL,WINEVETLOG --flagfile=config.cfg
      logstory usecase_replay_logtype RULES_SEARCH_WORKSHOP "POWERSHELL, WINEVETLOG" --flagfile=config.cfg

      # change loglevel and tee output to a file
      PYTHONLOGLEVEL=debug logstory ... | tee "logstory_exe_$(date +%Y%m%d_%H%M%S).log"

      # Entities
      logstory usecase_replay NETWORK_ANALYSIS --flagfile=config.cfg --timestamp_delta=1d --entities

      logstory usecase_replay_logtype NETWORK_ANALYSIS WIONDOWS_AD --flagfile=config.cfg --entities

      logstory usecase_get EDR_WORKSHOP
    """
  print(help_text)
  print(FLAGS)


def entry_point():
  if "--help" in sys.argv:
    custom_help()
    sys.exit(1)
  app.run(local_main)  # evaluates the flags


if __name__ == "__main__":
  if "--help" in sys.argv:
    custom_help()
    sys.exit(1)
  app.run(local_main)  # evaluates the flags



import yaml

def validate_base_time(filepath):
    """
    Reads a YAML file, checks each entry, and ensures it has exactly one
    base_time: true timestamp in its timestamps list.

    Args:
        filepath: The path to the YAML file.

    Raises:
        ValueError: If an entry has zero or more than one base_time: true timestamp.
    """
    try:
        with open(filepath, 'r') as f:
            data = yaml.safe_load(f)
    except FileNotFoundError:
        raise FileNotFoundError(f"File not found: {filepath}")
    except yaml.YAMLError as e:
        raise ValueError(f"Error parsing YAML file: {e}")

    if not isinstance(data, dict):
        raise ValueError("YAML file should contain a dictionary at the root level.")

    for entry_name, entry_data in data.items():
        if 'timestamps' not in entry_data:
            print(f"Warning: Entry '{entry_name}' has no 'timestamps' list. Skipping.")
            continue

        timestamps = entry_data['timestamps']
        base_time_count = 0

        for timestamp in timestamps:
            if 'base_time' in timestamp and timestamp['base_time']:
                base_time_count += 1

        if base_time_count == 0:
            raise ValueError(f"Entry '{entry_name}' has no base_time: true timestamp.")
        elif base_time_count > 1:
            raise ValueError(f"Entry '{entry_name}' has multiple base_time: true timestamps ({base_time_count}).")
        else:
            print(f"Entry '{entry_name}' has exactly one base_time: true timestamp. OK")


# Example usage (replace with the actual file path):
filepaths = [
    "../src/logstory/logtypes_entities_timestamps.yaml",
    "../src/logstory/logtypes_events_timestamps.yaml",
]
for filepath in filepaths:
  try:
      validate_base_time(filepath)
  except (ValueError, FileNotFoundError) as e:
      print(f"Error: {e}")



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logstory Events replay."""

import datetime
import json
import os
import re

from typing import Any, Dict, List, Optional

from google.auth.transport import requests
from google.cloud import secretmanager
from google.cloud import storage
from google.oauth2 import service_account

import requests as real_requests
import yaml


level = os.environ.get("PYTHONLOGLEVEL", "INFO").upper()
try:  # main.py shouldn't need abseil
  # pylint: disable-next=g-import-not-at-top
  from absl import logging as LOGGER

  absl_log_levels = {
      "DEBUG": LOGGER.DEBUG,
      "INFO": LOGGER.INFO,
      "WARNING": LOGGER.WARNING,
      "ERROR": LOGGER.ERROR,
      "FATAL": LOGGER.FATAL,
  }
  LOGGER.set_verbosity(absl_log_levels.get(level, LOGGER.INFO))
except ImportError:  # set up logging without abseil
  # pylint: disable-next=g-import-not-at-top
  import logging

  logging.basicConfig(level=getattr(logging, level, logging.INFO))
  logging.getLogger().setLevel(getattr(logging, level, logging.INFO))
  LOGGER = logging.getLogger(__name__)


# Constants common to all cloud functions
SCOPES = ["https://www.googleapis.com/auth/malachite-ingestion"]
SECRET_MANAGER_CREDENTIALS = os.environ.get("SECRET_MANAGER_CREDENTIALS")
CUSTOMER_ID = os.environ.get("CUSTOMER_ID")
CREDENTIALS_PATH = os.environ.get("CREDENTIALS_PATH")

REGION = os.environ.get("REGION")
url_prefix = f"{REGION}-" if REGION else ""
if url_prefix.lower().strip() == "us-":  # Region US is blank rather than "us-"
  url_prefix = ""

INGESTION_API_BASE_URL = (
    os.environ.get("INGESTION_API_BASE_URL")
    or f"https://{url_prefix}malachiteingestion-pa.googleapis.com"
)
# varies by cloud function
BUCKET_NAME = os.environ.get("BUCKET_NAME")
UTC = datetime.timezone.utc

storage_client = None
if os.getenv("SECRET_MANAGER_CREDENTIALS"):  # Running in a cloud function
  secretmanager_client = secretmanager.SecretManagerServiceClient()
  storage_client = storage.Client()
  sec_request = {"name": f"{SECRET_MANAGER_CREDENTIALS}/versions/latest"}
  sec_response = secretmanager_client.access_secret_version(sec_request)
  ret = sec_response.payload.data.decode("UTF-8")
  service_account_info = json.loads(ret)
else:
  # Running locally (not in cloud function)
  if CREDENTIALS_PATH:  # optional to facilitate unit testing
    with open(CREDENTIALS_PATH) as f:
      service_account_info = json.load(f)
  else:
    service_account_info = None

# Create a credential using SA Credential and Ingestion API Scope.
if service_account_info:  # optional to facilitate unit testing
  credentials = service_account.Credentials.from_service_account_info(
      service_account_info, scopes=SCOPES
  )

  # Build an HTTP client which can make authorized OAuth requests.
  http_client = requests.AuthorizedSession(credentials)


def _get_timestamp_delta_dict(timestamp_delta: str) -> Dict[str, int]:
  """Parses the timestamp delta string into a dictionary."""
  ts_delta_pairs = re.findall(r"(\d+)([dhm])", timestamp_delta)
  return {letter: int(number) for number, letter in ts_delta_pairs}


def _get_log_content(
    use_case: str, log_type: str, entities: Optional[bool] = False
) -> str:
  """Retrieves log content from either GCS or local filesystem."""
  if entities:
    object_name = f"{use_case}/ENTITIES/{log_type}.log"
  else:
    object_name = f"{use_case}/EVENTS/{log_type}.log"

  LOGGER.info("Processing file: %s", object_name)
  if storage_client:  # running in cloud function
    bucket = storage_client.bucket(BUCKET_NAME)
    file_object = bucket.get_blob(object_name)
    return file_object.download_as_text()
  else:
    # Local filesystem case
    script_dir = os.path.dirname(os.path.abspath(__file__))
    # local_file_path = os.path.join(script_dir, "../../usecases/", object_name)
    local_file_path = os.path.join(script_dir, "usecases/", object_name)
    with open(local_file_path, "r") as f:
      return f.read()


def _get_ingestion_labels(
    use_case: str,
    logstory_exe_time: datetime.datetime,
    api_for_log_type: str,
) -> List[Dict[str, Any]]:
  """Constructs the ingestion labels list."""
  return [
      {
          "key": "ingestion_method",
          "value": api_for_log_type,
      },
      {
          "key": "replayed_from",
          "value": "logstory",
      },
      {
          "key": "source_usecase",
          "value": use_case,
      },
      {
          "key": "log_replay",
          "value": "true",
      },
      {
          "key": "log_replay_time",
          # changes for each logtype in each usecase
          "value": _get_current_time().isoformat(),
      },
      {
          "key": "logstory_exe_time",
          # same value for all logtypes in all usecases. i.e. 1 value per run.
          "value": logstory_exe_time.isoformat(),
      },
  ]


def _get_current_time():
  """Returns the current time in UTC."""
  return datetime.datetime.now(UTC)


def _update_timestamp(
    log_text: str,
    timestamp: Dict[str, str],
    old_base_time: datetime.datetime,
    ts_delta_dict: Dict[str, int],
) -> str:
  """Updates 1 timestamp in the line of log text based on provided parameters.

  Args:
    log_text: string containing timestamp and other text
    timestamp: describes the timestamp pattern to search for
    old_base_time: the first ts in the first line of the first logfile
    ts_delta_dict: user configured offset from current datetime
      the updated timestamp will be now() - [Nd]days -[Nh]hours - [Nm]mins
      ex. 2024-03-14T13:37:42.123456Z and {d: 1, h: 1, m: 1} ->
          2024-03-13T12:36:42.123456Z
      Note that the seconds and miliseconds are always preserved.

  Returns:
    log_text: updated timestamp string
  """
  group_index = int(timestamp["group"]) - 1
  ts_match = re.search(timestamp["pattern"], log_text)
  if ts_match:
    # Return value is a tuple, cast it to list, so we can replace one item
    groups = list(ts_match.groups())
    event_timestamp = groups[group_index]
    event_time = None
    if timestamp.get("epoch"):
      dateformat = "%s"
      event_time = datetime.datetime.fromtimestamp(int(event_timestamp))
    elif timestamp.get("pattern"):
      dateformat = str(timestamp["dateformat"])
      event_time = datetime.datetime.strptime(event_timestamp, dateformat)
    else:
      # Raise?
      return log_text

    if event_time:
      # `old_base_time` is the base t (bts) in the first line of the
      #  first logfile of the usecase's set of logfiles.
      # If the current timestamp has a different date from the old_base_time
      #  we want the same N days different to be in the final ts

      if event_time.year == 1900:
        event_time = event_time.replace(year=old_base_time.year)
      more_days = (event_time.date() - old_base_time.date()).days
      # Get today's date minus N days
      subtract_n_days = ts_delta_dict.get("d", 0) - more_days
      new_day = _get_current_time() - datetime.timedelta(days=subtract_n_days)
      # Update date but keep the original time
      new_event_time = event_time.replace(
          year=new_day.year,
          month=new_day.month,
          day=new_day.day,
      )
      # now update the time if user provided [Nh][Nm]
      # the optional h/m delta enables running > 1x/day
      if "h" in ts_delta_dict or "m" in ts_delta_dict:
        hm_delta = datetime.timedelta(
            hours=ts_delta_dict.get("h", 0),
            minutes=ts_delta_dict.get("m", 0),
        )
        new_event_time = new_event_time - hm_delta
      new_event_timestamp = new_event_time.strftime(dateformat)
      groups[group_index] = new_event_timestamp
      log_text = re.sub(timestamp["pattern"], "".join(groups), log_text)
  return log_text


def _post_entries_in_batches(
    api: str,
    log_type: str,
    all_entries: List[Dict[str, str]],
    ingestion_labels: List[Dict[str, str]],
):
  """Posts entries to the ingestion API in batches."""
  entries_bytes = 0
  entries = []
  for i, entry in enumerate(all_entries):
    entries.append(entry)
    try:
      entries_bytes += len(entry["logText"])  # unstructuredlogentry
    except (TypeError, KeyError):
      entries_bytes += len(entry)  # udmevent
    if len(entries) % 1000 == 0 or entries_bytes > 500_000:
      LOGGER.info("posting entry N: %s, entries_bytes: %s", i, entries_bytes)
      post_entries(api, log_type, entries, ingestion_labels)
      entries = []
      entries_bytes = 0

  # after the loop, also submit if there are leftover entries
  if entries:
    LOGGER.info("posting remaining entries")
    post_entries(api, log_type, entries, ingestion_labels)


# pylint: disable-next=g-bare-generic
def post_entries(api: str, log_type: str, entries: list, ingestion_labels: dict):
  """Send the provided entries to the appropriate ingestion API method.

  Args:
    api: which ingestion API to use (unstructured or udm)
    log_type: value from yaml; unused if UDM
    entries: list of entries to send to ingestion API
    ingestion_labels:
  Returns:
    None
  """
  if api == "unstructuredlogentries":
    uri = f"{INGESTION_API_BASE_URL}/v2/unstructuredlogentries:batchCreate"
    body = json.dumps(
        {
            "customer_id": CUSTOMER_ID,
            "log_type": log_type,
            "entries": entries,
            "labels": ingestion_labels,
        }
    )
    response = http_client.post(uri, data=body)
  elif api == "udmevents":
    for entry in entries:
      if "metadata" not in entry:
        entry["metadata"] = {}
      entry["metadata"]["ingestion_labels"] = ingestion_labels

    uri = f"{INGESTION_API_BASE_URL}/v2/udmevents:batchCreate"
    data = {
        "customer_id": CUSTOMER_ID,
        "events": entries,
    }
    response = http_client.post(uri, json=data)
  elif api == "entities":
    uri = f"{INGESTION_API_BASE_URL}/v2/entities:batchCreate"
    body = json.dumps(
        {
            "customer_id": CUSTOMER_ID,
            "log_type": log_type,
            "entities": entries,
        }
    )
    response = http_client.post(uri, data=body)

  try:
    response.raise_for_status()
  except real_requests.exceptions.HTTPError as err:
    try:
      response_data = response.json()  # Try to parse JSON response
    except ValueError:
      response_data = response.text  # Fallback to raw text
    LOGGER.error("Response content: %s", response_data)
    # Re-raise the exception to propagate it
    raise err
  LOGGER.info("Response Status Code: %s", response.status_code)


# pylint: disable-next=missing-function-docstring
def usecase_replay_logtype(
    use_case: str,
    log_type: str,
    logstory_exe_time: datetime.datetime,
    old_base_time: Optional[datetime.datetime] = None,
    timestamp_delta: Optional[str] = None,
    ts_map_path: Optional[str] = "./",
    entities: Optional[bool] = False,
) -> Optional[datetime.datetime]:
  """Replays log data for a specific use case and log type.

  Args:
    use_case: value from yaml (ex. AWS, AZURE_AD, ...)
    log_type: value from yaml (ex. CS_EDR, GCP_CLOUDAUDIT, ...)
    logstory_exe_time: common to all logtypes and all usecases
    old_base_time: the first base timestamp in the first line of the first log.
    timestamp_delta: [Nd][Nh][Nm] string, for calculating the timestampdelta
     to apply to each timestamp pattern match.
    ts_map_path: disk location of the yaml files
    entities: bool for Entities (True) vs Events (False)

  Returns:
    old_base_time: so that subequent logtypes/usecases can all use the same value
  """
  timestamp_delta = timestamp_delta or "1d"
  ts_delta_dict = _get_timestamp_delta_dict(timestamp_delta)

  # Construct the full path to the YAML file
  if entities:
    file_path = os.path.join(ts_map_path, "logtypes_entities_timestamps.yaml")
  else:
    file_path = os.path.join(ts_map_path, "logtypes_events_timestamps.yaml")

  if file_path.startswith("."):
    file_path = os.path.split(__file__)[0] + "/" + file_path
  with open(file_path) as fh:
    timestamp_map = yaml.safe_load(fh)
    api_for_log_type = timestamp_map[log_type]["api"]
    log_content = _get_log_content(use_case, log_type, entities)
    ingestion_labels = _get_ingestion_labels(
        use_case, logstory_exe_time, api_for_log_type
    )
    # base time stamp (BTS) determines the anchor point; others are relative
    btspattern, btsgroup, btsformat, btsepoch = [
        (
            timestamp.get("pattern"),
            timestamp.get("group"),
            timestamp.get("format"),
            timestamp.get("epoch"),
        )
        for timestamp in timestamp_map[log_type]["timestamps"]
        if timestamp.get("base_time")
    ][0]
    entries = []
    for line_no, log_text in enumerate(log_content.splitlines()):
      if old_base_time is None:
        old_base_match = re.search(btspattern, log_text)
        if old_base_match and old_base_match.groups():
          old_base = old_base_match.group(btsgroup)
          if btsepoch:
            old_base_time = datetime.datetime.fromtimestamp(int(old_base))
          elif btsformat:
            old_base_time = datetime.datetime.strptime(old_base, btsformat)

      # for each timestamp in the yaml file
      for ts_n, timestamp in enumerate(timestamp_map[log_type]["timestamps"]):
        log_text = _update_timestamp(
            log_text,
            timestamp,
            old_base_time,
            ts_delta_dict,
        )
        LOGGER.debug("Finished processing line N: %s, timestamp N: %s", line_no, ts_n)

      # accumulate all of the entries into memory
      LOGGER.debug("log_text after all ts updates: %s", log_text)
      LOGGER.debug("now as repr:")
      print(repr(log_text))
      if api_for_log_type == "unstructuredlogentries":
        entries.append({"logText": log_text})
      elif api_for_log_type == "udmevents":
        entries.append(json.loads(log_text))
      elif api_for_log_type == "entities":
        entries.append(json.loads(log_text))
      else:
        raise ValueError("Only unstructuredlogentries and udmevents are supported")

    _post_entries_in_batches(
        api_for_log_type,
        log_type,
        entries,
        ingestion_labels,
    )
  return old_base_time


def main(request=None, enabled=False):  # pylint: disable=unused-argument
  """Read config and call usecase_replay_logtype for each [usecase, logtype].

  This is the entry point for the Cloud Function. It is not used by CLI.

  It calls usecases_replay for each of the usecases that is
   in the file usecases_events_logtype_map.yaml where: enabled: 1
  There is also log_type list in that config. That is duplicative if all of
   the log types in the subdir are used but it is useful to be able to toggle
   some off.

  old_base_time is the first base timestamp in the first line of the first log.
   - all changes are relative to that value: both interlog and intraline.
   - if line N of log N is > than old_base_time, the event_time will be in the future.
   - if it is > 30 days in the future, it will be 9999-01-01

  NOTE: this fx is not used by logstory.py, so local filenames are ok

  Args:
   request: Unused; Needed for running in Cloud Function.
   enabled: Flag to force running all usecases.

  Returns:
   Success Events!
  """
  logstory_exe_time = _get_current_time()
  entities = os.getenv("ENTITIES")
  if entities:
    filename = "usecases_entities_logtype_map.yaml"
  else:
    filename = "usecases_events_logtype_map.yaml"

  with open(os.path.join(os.path.dirname(__file__), filename)) as fh:
    yaml_use_cases = yaml.safe_load(fh)
    use_cases = list(yaml_use_cases.keys())
    for use_case in use_cases:
      if (yaml_use_cases[use_case]["enabled"]) > 0 or enabled:
        LOGGER.info("use_case: %s", use_case)
        old_base_time = None  # reset for each usecase
        for log_type in yaml_use_cases[use_case]["log_type"]:
          old_base_time = None  # reset for each log_type
          LOGGER.info("log_type: %s", log_type)
          old_base_time = usecase_replay_logtype(
              use_case,
              log_type,
              logstory_exe_time,
              old_base_time,
              timestamp_delta="1d",
              entities=entities,  # bool
          )
    LOGGER.info("use_case: %s completed.", use_case)
    logstory_fin_time = _get_current_time()
    LOGGER.info(  # pylint: disable=logging-fstring-interpolation
        f"""UDM Search for the loaded logs:
    metadata.ingested_timestamp.seconds >= {int(logstory_exe_time.timestamp())}
    metadata.log_type = {log_type}
    metadata.ingestion_labels["replayed_from"] = "logstory"
    metadata.ingestion_labels["log_replay"] = "true"
    metadata.ingestion_labels["usecase_name"] = "{use_case}"
    """
    )
  return "Success Events!"
